/*******************************************************************************
 
Autore : Mauro Laurenti
Versione : 1.2
 
Created on Date : 21/07/2019
Last update     : 15/05/2024

CopyRight 2006-2023 all rights are reserved

********************************************************
SOFTWARE LICENSE AGREEMENT
********************************************************

The usage of the supplied software imply the acceptance of the following license.

The software supplied herewith by Mauro Laurenti (the Author) is intended for
use solely and exclusively on Microchip PIC Microcontroller (registered mark).
The software is owned by the Author, and is protected under applicable
copyright laws. All rights are reserved.
Any use in violation of the foregoing restrictions may subject the
user to criminal sanctions under applicable laws, as well as to civil liability
for the breach of the terms and conditions of this license.
Commercial use is forbidden without a written acknowledgement with the Author.
Personal or educational use is allowed if the application containing the
following software doesn't aim to commercial use or monetary earning of any kind.

THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE AUTHOR SHALL NOT,
IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 
*******************************************************************************/

#ifdef COMPILER_XC
    #include <xc.h>
#endif

#ifndef LTLIB_UTILITIES_H
#define LTLIB_UTILITIES_H

//It sets how big a float conversion could be 
#define MAX_NUM_DIGITS 12

#ifndef DECIMAL_SEPARATOR 
    #define DECIMAL_SEPARATOR '.'
#endif

const uint8_t hex_table [] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

// ASCII Special codes 0-32
const uint8_t ASCII_Null = 0;
const uint8_t ASCII_Start_of_Heading = 1;
const uint8_t ASCII_Start_of_Text = 2;
const uint8_t ASCII_End_of_Text = 3;
const uint8_t ASCII_End_of_Transmit = 4;
const uint8_t ASCII_Enquiry = 5;
const uint8_t ASCII_Aknowledge = 6;
const uint8_t ASCII_Audible_Bell = 7;
const uint8_t ASCII_Back_Space = 8;
const uint8_t ASCII_Horizontal_Tab = 9;
const uint8_t ASCII_New_Line = 10;  //Two Name definitions for easy of use 
const uint8_t ASCII_Line_Feed = 10; //Two Name definitions for easy of use
const uint8_t ASCII_Vertical_Tab = 11;
const uint8_t ASCII_Form_Feed = 12;
const uint8_t ASCII_Carriage_Return = 13;
const uint8_t ASCII_Shift_Out = 14;
const uint8_t ASCII_Shift_In = 15;
const uint8_t ASCII_Data_Link_Escape = 16;
const uint8_t ASCII_Device_Control_1 = 17;
const uint8_t ASCII_Device_Control_2 = 18;
const uint8_t ASCII_Device_Control_3 = 19;
const uint8_t ASCII_Device_Control_4 = 20;
const uint8_t ASCII_Neg_Aknowledge = 21;
const uint8_t ASCII_Synchronous_Idle = 22;
const uint8_t ASCII_End_Transmission_Block = 23;
const uint8_t ASCII_Cancel = 24;
const uint8_t ASCII_End_Of_Medium = 25;
const uint8_t ASCII_Substitution = 26;
const uint8_t ASCII_Escape = 27;
const uint8_t ASCII_File_Separator = 28;
const uint8_t ASCII_Group_Separator = 29;
const uint8_t ASCII_Record_Separator = 30;
const uint8_t ASCII_Unit_Separator = 31;
const uint8_t ASCII_Space = 32;


/**
 * This function converts signed integer to a string of chars.
 * The returned string is always the shortest possible one containing the value.
 * Digit 0 contains - sign if the number is negative. The sign + is not shown.
 *
 * @param buffer Pointer to the array that will contain the string.
 *
 * @param value Integer value to convert
 * 
 * @return error 0: Conversion OK
 *              -1: Overflow - The number is too big  
 */
int8_t LTlib_int16_to_strings (uint8_t * buffer, int16_t value);


/**
 * This function converts a float number to a string of chars.
 *
 * @param buffer buffer Pointer to the array that will contain the string.
 *
 * @param value float value to convert (float)
 * 
 * @param fractional_digits it specified the number of fractional digits 
 * 
 * @return error 0: Conversion OK
 *              -1: Overflow - The number is too big.  
 * 
 * @warning The buffer must be at least of size 8 to avoid overflow. 
 *          The function generates an internal buffer of MAX_NUM_DIGITS size.
 */
int8_t LTlib_float_to_string (uint8_t * buffer, float value,uint8_t fractional_digits );


/**
 * This function converts unsigned integer (HEX 8) to a string of chars made of ASCII digits.
 *
 * @param buffer buffer Pointer to the array that will contain the string.
 *
 * @param value Integer value to convert (uint8_t)
 * 
 * @return error 0: Conversion OK
 *              -1: Overflow - The number is too big.  
 * 
 * @warning The buffer must be at least of size 5 to avoid overflow.
 */
int8_t LTlib_Hex8_to_string (uint8_t * buffer, uint8_t value);


/**
 * This function converts unsigned integer (HEX 16) to a string of chars made of ASCII digits.
 *
 * @param buffer buffer Pointer to the array that will contain the string.
 *
 * @param value Integer value to convert (uint16_t)
 * 
 * @return error 0: Conversion OK
 *              -1: Overflow - The number is too big.  
 * 
 * @warning The buffer must be at least of size 5 to avoid overflow.
 */
int8_t LTlib_Hex16_to_string (uint8_t * buffer, uint16_t value);


/**
 * This function converts unsigned integer (HEX 32) to a string of chars made of HEX digits.
 *
 * @param buffer buffer Pointer to the array that will contain the string.
 *
 * @param value Integer value to convert (uint32_t)
 * 
 * @return error 0: Conversion OK
 *              -1: Overflow - The number is too big.  
 * 
 * @warning The buffer must be at least of size 8 to avoid overflow.
 */
int8_t LTlib_Hex32_to_string (uint8_t * buffer, uint32_t value);


/**
 * This function converts unsigned integer (BCD 8) to a uint8_t (integer value of the BCD code).
 *
 * @param value BCD value to convert (uint8_t)
 * 
 * @return result
 *
 */
uint8_t LTlib_BCD8_to_uint8 (uint8_t value);


/**
 * This function returns the string length, marked with the '\0' character
 * The \0 character is not counted.
 *
 * @param buffer Pointer to the string that contains the char array ending with '\0'
 * 
 * @return value String length 
 *      
 */
uint8_t LTlib_string_length (uint8_t * buffer);


/**
 * This function returns the string length, marked with the '\0' character
 * The \0 character is not counted.
 *
 * @param string_1 Pointer to the string 1 to be merged that contains the char array ending with '\0'
 * 
 * @param string_2 Pointer to the string 1 to be merged that contains the char array ending with '\0'
 * 
 * @param string_merged Pointer to the string 1 to be merged that contains the char array ending with '\0'
 * 
 * @return value String length 
 *      
 */
void LTlib_merge_string (uint8_t * string_1, uint8_t * string_2, uint8_t * string_merged);



#endif

 
