/*******************************************************************************
 
Modified by: Mauro Laurenti
Modifications: New functions, file organization, comments and naming. 
               The functions interface to the FatFs library remains the same.

Version : 1.0 based on FatFs 1.3c
 
Created on Date : 09/05/2024
Last update     : 09/05/2024

CopyRight 2006-2024 all rights are reserved

********************************************************
SOFTWARE LICENSE AGREEMENT
********************************************************

The Library is based on the ChaN's FatFs Library which owns the copyright, under 
which provides the following rights:  
   
------------------------------------------------------------------------
  MMCv3/SDv1/SDv2 (in SPI mode) control module
-------------------------------------------------------------------------

Copyright (C) 2010, ChaN, all right reserved.

This software (FatFs Library) is a free software and there is NO WARRANTY.
No restriction on use. You can use, modify and redistribute it for
personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
Redistributions of source code must retain the above copyright notice.

/------------------------------------------------------------------------- 
 
The usage of the supplied software,as part of LTlib, implies the acceptance of the following license.

The software supplied herewith by Mauro Laurenti (Author of the modified version)
is owned by the Author, and is protected under applicable copyright laws. 
All rights are reserved except for the FatFs Library itself that can be further used
according to ChaN's license.  
Any use in violation of the foregoing restrictions may subject the
user to criminal sanctions under applicable laws, as well as to civil liability
for the breach of the terms and conditions of this license.
Commercial use is forbidden without a written acknowledgement with the Author.
Personal or educational use is allowed if the application containing the
following software doesn't aim to commercial use or monetary earning of any kind.

THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE AUTHOR SHALL NOT,
IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 
*******************************************************************************/

#ifdef COMPILER_XC
    #include <xc.h>
#endif

#ifndef _DISKIO_DEFINED
#define _DISKIO_DEFINED

#ifdef __cplusplus
extern "C" {
#endif

    
//*****************************************
//   Library Test
//*****************************************
#include "LTlib.h"

#ifndef LTLIB_SPI_MODULE_H
    #error The memory_raw Library requires the "module_SPI" library to be included.
#endif
    
#ifndef LTLIB_SPI_MODULE_H
    #error The memory_raw Library requires the "module_SPI" library to be included.
#endif    

#ifdef MEMORY_BOARD_WITH_RTC
    #ifndef LTLIB_RTC
    #error The memory_raw Library requires an RTC library to be included (e.g.MCP7940.h).
    #endif   

    #ifndef LTLIB_UTILITIES_H
    #error The memory_raw Library by using the RTC library, requires LTlib_utilities library.
    #endif   
    
    
#endif    

//*****************************************
//   SPI Module Switch
//*****************************************
#if defined (MEMORY_SPI_MODULE_1) || defined (MEMORY_SPI_MODULE_2)

#else
    #define MEMORY_SPI_MODULE_1
#endif 


#ifdef MEMORY_SPI_MODULE_1
    #define MEMORY_SPI_open SPI1_open
    #define MEMORY_SPI_close SPI1_close
    #define MEMORY_SPI_baudrate SPI1_baudrate
    #define MEMORY_SPI_mode SPI1_mode 
    #define MEMORY_SPI_write_byte SPI1_write_byte
    #define MEMORY_SPI_read_byte SPI1_read_byte
#endif

#ifdef MEMORY_SPI_MODULE_2 
    #define MEMORY_SPI_open SPI2_open
    #define MEMORY_SPI_close SPI2_close
    #define MEMORY_SPI_baudrate SPI2_baudrate
    #define MEMORY_SPI_mode SPI2_mode
    #define MEMORY_SPI_write_byte SPI2_write_byte
    #define MEMORY_SPI_read_byte SPI2_read_byte
#endif

//**************************************************
// Device pin Settings CS
//**************************************************

#ifdef COMPILER_XC8
    #ifdef _PIC18
        //CS pin
        #define MEMORY_CS_TRIS     TRISAbits.RA5
        #define MEMORY_CS_PORT     LATAbits.LATA5
    #endif

    #ifndef _PIC18
        //CS pin
        #define MEMORY_CS_TRIS     TRISAbits.TRISA5
        #define MEMORY_CS_PORT     PORTAbits.RA5
    #endif
#endif

#ifdef COMPILER_XC16
        //CS pin
        #define MEMORY_CS_TRIS    TRISCbits.TRISC2
        #define MEMORY_CS_PIN     LATCbits.LATC2

        //Card Detection
        #define MEMORY_DETECTION_TRIS    TRISCbits.TRISC5
        #define MEMORY_DETECTION_PIN     PORTCbits.RC5
#endif       
    
    
#ifdef COMPILER_XC32
        //CS pin
        #define MEMORY_CS_TRIS    TRISCbits.TRISC2
        #define MEMORY_CS_PIN     LATCbits.LATC2

        //Card Detection
        #define MEMORY_DETECTION_TRIS    TRISCbits.TRISC5
        #define MEMORY_DETECTION_PIN     PORTCbits.RC5
#endif    


//**************************************************
// SPI Speed
//**************************************************

#define MEMORY_LOW_SPEED SPI_CLK_50_KHZ
#define MEMORY_HIGH_SPEED SPI_CLK_1000_KHZ
    

//**************************************************
// Constants 
//**************************************************
#define MEMORY_OUTPUT_PIN  0x00  
#define MEMORY_INPUT_PIN   0x01
    
#define MEMORY_CS_PIN_DISABLED 0x01
#define MEMORY_CS_PIN_ENABLED  0x00
    
#define MEMORY_DETECTED 0x01
#define MEMORY_NOT_DETECTED  0x00    
    

    
//**************************************************    

/* Status of Disk Functions */
typedef BYTE	DSTATUS;

/* Results of Disk Functions */
typedef enum {
	RES_OK = 0,		/* 0: Successful */
	RES_ERROR,		/* 1: R/W Error */
	RES_WRPRT,		/* 2: Write Protected */
	RES_NOTRDY,		/* 3: Not Ready */
	RES_PARERR		/* 4: Invalid Parameter */
} DRESULT;


/* Disk Status Bits (DSTATUS) */

#define STA_NOINIT		0x01	/* Drive not initialized */
#define STA_NODISK		0x02	/* No medium in the drive */
#define STA_PROTECT		0x04	/* Write protected */


/* Command code for disk_ioctrl function */
#define _READONLY 0
/* Generic command (Used by FatFs) */
#define CTRL_SYNC			0	/* Complete pending write process (needed at FF_FS_READONLY == 0) */
#define GET_SECTOR_COUNT	1	/* Get media size (needed at FF_USE_MKFS == 1) */
#define GET_SECTOR_SIZE		2	/* Get sector size (needed at FF_MAX_SS != FF_MIN_SS) */
#define GET_BLOCK_SIZE		3	/* Get erase block size (needed at FF_USE_MKFS == 1) */
#define CTRL_TRIM			4	/* Inform device that the data on the block of sectors is no longer used (needed at FF_USE_TRIM == 1) */

/* Generic command (Not used by FatFs) */
#define CTRL_POWER			5	/* Get/Set power status */
#define CTRL_LOCK			6	/* Lock/Unlock media removal */
#define CTRL_EJECT			7	/* Eject media */
#define CTRL_FORMAT			8	/* Create physical format on the media */

/* MMC/SDC specific ioctl command */
#define MMC_GET_TYPE		10	/* Get card type */
#define MMC_GET_CSD			11	/* Get CSD */
#define MMC_GET_CID			12	/* Get CID */
#define MMC_GET_OCR			13	/* Get OCR */
#define MMC_GET_SDSTAT		14	/* Get SD status */
#define ISDIO_READ			55	/* Read data form SD iSDIO register */
#define ISDIO_WRITE			56	/* Write data to SD iSDIO register */
#define ISDIO_MRITE			57	/* Masked write data to SD iSDIO register */

/* ATA/CF specific ioctl command */
#define ATA_GET_REV			20	/* Get F/W revision */
#define ATA_GET_MODEL		21	/* Get model name */
#define ATA_GET_SN			22	/* Get serial number */

/* MMC card type flags (MMC_GET_TYPE) */
#define CT_MMC		0x01		/* MMC ver 3 */
#define CT_SD1		0x02		/* SD ver 1 */
#define CT_SD2		0x04		/* SD ver 2 */
#define CT_SDC		(CT_SD1|CT_SD2)	/* SD */
#define CT_BLOCK	0x08		/* Block addressing */



/**
 * This function initializes the Disk Drive
 *
 * @param pdrv pointer to the driver. Physical drive number
 * 
 * @return status STA_NOINIT Drive not initialized 
 *               STA_NODISK No medium in the drive  
 *               STA_PROTECT Write protected
 */
DSTATUS disk_initialize (BYTE pdrv);


/**
 * This function gets the Disk Status 
 *
 * @param pdrv pointer to the driver. Physical drive number
 * 
 * @return status STA_NOINIT Drive not initialized 
 *               STA_NODISK No medium in the drive  
 *               STA_PROTECT Write protected
 */
DSTATUS disk_status (BYTE pdrv);


/**
 * This function reads the disk sector/s
 *
 * @param pdrv pointer to the driver. Physical drive number
 * 
 * @param buff pointer to the buffer where to store the data. 
 * 
 * @param sector Start sector from which read the data.
 * 
 * @param count Number of sectors to read.
 * 
 * @return result RES_OK Successful
 *                RES_ERROR  R/W Error 
 *                RES_WRPRT Write Protected
 *                RES_NOTRDY Not Ready
 *                RES_PARERR Invalid Parameter
 */

DRESULT disk_read (BYTE pdrv, BYTE* buff, DWORD sector, UINT count);


/**
 * This function writes the disk sector/s
 *
 * @param pdrv pointer to the driver. Physical drive number
 * 
 * @param buff pointer to the buffer where the data to write is stored. 
 * 
 * @param sector Start sector from which write the data.
 * 
 * @param count Number of sectors to read.
 * 
 * @return result RES_OK Successful
 *                RES_ERROR  R/W Error 
 *                RES_WRPRT Write Protected
 *                RES_NOTRDY Not Ready
 *                RES_PARERR Invalid Parameter
 */
DRESULT disk_write (BYTE pdrv, const BYTE* buff, DWORD sector, UINT count);

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/
/* drv Physical drive number (0) */
/* ctrl Control code */
/* buff Buffer to send/receive data block */

/**
 * Miscellaneous Functions, selected via a command. 
 *
 * @param pdrv pointer to the driver. Physical drive number
 * 
 * @param cmd command to be executed. 
 * 
 * @param buff data buffer
 * 
 * @return result RES_OK Successful
 *                RES_ERROR  R/W Error 
 *                RES_WRPRT Write Protected
 *                RES_NOTRDY Not Ready
 *                RES_PARERR Invalid Parameter
 */
DRESULT disk_ioctl (BYTE pdrv, BYTE cmd, void* buff);


/**
 * This function waits till the memory is ready but with a timeout. 
 *
 * @param void 
 * 
 * @return void
 */
static BYTE wait_memory_ready (void);


/**
 * This function deactivates the memory and SPI module
 *
 * @param void 
 * 
 * @return void
 */
static void deactivate_memory (void);


/**
 * This function activates the memory and SPI module
 *
 * @param void 
 * 
 * @return void
 */
static void activate_memory (void);


/**
 * This function deactivates the CS associated to the SPI module, driving the memory.
 * It will disable the memory R/W process.
 *
 * @param void 
 * 
 * @return void
 */
static void memory_CS_disabled (void);


/**
 * This function activate the CS associated to the SPI module, driving the memory.
 * It will enable the memory R/W process. 
 * Upon memory selection, it waits for the card to be ready. 
 * 
 * @param void 
 * 
 * @return 1: successful 0: Timeout
 */
static int memory_CS_enabled (void);


/**
 * This function writes a command to the memory. 
 * 
 * @param cmd Command to be sent 
 * 
 * @param arg additional data associated to the command 
 * 
 * @return error code
 */
static BYTE send_cmd ( BYTE cmd, DWORD arg);


/**
 * This function receives data packet from the MMC memory.
 * 
 * @param buff Buffer where to store the received data. 
 * 
 * @param btr Byte count 
 * 
 * @return error code 1:OK 0:Failed
 */
static int rcvr_datablock (	BYTE *buff, UINT btr );


/**
 * This function check if the memory is connected by checking the inserted switch. 
 * 
 * @param void
 * 
 * @return status code  MEMORY_DETECTED 
 *                      MEMORY_NOT_DETECTED 
 */
uint8_t is_memory_inserted (void);


/**
 * This function blinks the LED on the memory board. The LED is turned ON for the defined
 * time and than it is turned back OFF. The LED may not be present on all boards. 
 * 
 * @param time_ON It defined how many ms the LED will remain ON [1...255]
 * 
 * @return void
 */
void blink_memory_LED(uint8_t time_ON);


/**
 * This function is called by FATfs to retrieve the time to be used for the file.
 * MEMORY_BOARD_WITH_RTC must be  defined before memory_raw.h to enable RTC support.
 * FF_FS_NORTC, within FAT.h is automatically updated with the right value. 
 * 
 * @param void
 * 
 * @return DWORD Time and date out of the RTC
 * 
 * @note 
 *          bit31:25     Year origin from the 1980 (0..127, e.g. 37 for 2017)
 *          bit24:21     Month (1..12)
 *          bit20:16     Day of the month (1..31)
 *          bit15:11     Hour (0..23)
 *          bit10:5      Minute (0..59)
 *          bit4:0       Second / 2 (0..29, e.g. 25 for 50) 
 * 

 */

#ifdef MEMORY_BOARD_WITH_RTC    
DWORD get_fattime (void);
#endif

#ifdef __cplusplus
}
#endif

#endif
