/*******************************************************************************
 
Author : Mauro Laurenti
Modifications: New functions, file organization, comments and naming. 
               The functions interface to the FatFs library remains the same.
 
Version : 1.0 based on FatFs 1.3c


Created on Date : 09/05/2024
Last update     : 09/05/2024

CopyRight 2006-2024 all rights are reserved

********************************************************
SOFTWARE LICENSE AGREEMENT
********************************************************

The Library is based on the ChaN's FatFs Library which owns the copyright, under 
which provides the following rights:  
   
------------------------------------------------------------------------
  MMCv3/SDv1/SDv2 (in SPI mode) control module
-------------------------------------------------------------------------

Copyright (C) 2010, ChaN, all right reserved.

This software (FatFs Library) is a free software and there is NO WARRANTY.
No restriction on use. You can use, modify and redistribute it for
personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
Redistributions of source code must retain the above copyright notice.

/------------------------------------------------------------------------- 
 
The usage of the supplied software,as part of LTlib, implies the acceptance of the following license.

The software supplied herewith by Mauro Laurenti (Author of the modified version)
is owned by the Author, and is protected under applicable copyright laws. 
All rights are reserved except for the FatFs Library itself that can be further used
according to ChaN's license.  
Any use in violation of the foregoing restrictions may subject the
user to criminal sanctions under applicable laws, as well as to civil liability
for the breach of the terms and conditions of this license.
Commercial use is forbidden without a written acknowledgement with the Author.
Personal or educational use is allowed if the application containing the
following software doesn't aim to commercial use or monetary earning of any kind.

THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE AUTHOR SHALL NOT,
IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 
*******************************************************************************/

#ifdef COMPILER_XC
    #include <xc.h>
#endif

#include "memory_raw.h"

#define _DISABLE_OPENADC10_CONFIGPORT_WARNING
#define _SUPPRESS_PLIB_WARNING

/* Definitions for MMC/SDC command */
#define CMD0   (0)			/* GO_IDLE_STATE */
#define CMD1   (1)			/* SEND_OP_COND */
#define ACMD41 (41|0x80)	/* SEND_OP_COND (SDC) */
#define CMD8   (8)			/* SEND_IF_COND */
#define CMD9   (9)			/* SEND_CSD */
#define CMD10  (10)			/* SEND_CID */
#define CMD12  (12)			/* STOP_TRANSMISSION */
#define ACMD13 (13|0x80)	/* SD_STATUS (SDC) */
#define CMD16  (16)			/* SET_BLOCKLEN */
#define CMD17  (17)			/* READ_SINGLE_BLOCK */
#define CMD18  (18)			/* READ_MULTIPLE_BLOCK */
#define CMD23  (23)			/* SET_BLOCK_COUNT */
#define ACMD23 (23|0x80)	/* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24  (24)			/* WRITE_BLOCK */
#define CMD25  (25)			/* WRITE_MULTIPLE_BLOCK */
#define CMD41  (41)			/* SEND_OP_COND (ACMD) */
#define CMD55  (55)			/* APP_CMD */
#define CMD58  (58)			/* READ_OCR */

static volatile DSTATUS Stat = STA_NOINIT;	/* Disk status */

static volatile UINT Timer1, Timer2;		/* 1000Hz decrement timer */

static UINT CardType;


//************************************************************
//          activate_memory  Implementation
//************************************************************
static void activate_memory (void){
    
    //Set the CS pin as Output
    MEMORY_CS_TRIS = MEMORY_OUTPUT_PIN;
       
    //Set the Detection pin as Input
    MEMORY_DETECTION_TRIS = MEMORY_INPUT_PIN;
          
    MEMORY_CS_PIN = MEMORY_CS_PIN_DISABLED;
    
    delay_ms (100);
        
    //SPI module initialization
    MEMORY_SPI_close();
    
    MEMORY_SPI_mode(SPI_MODE_3);
    MEMORY_SPI_baudrate(MEMORY_LOW_SPEED);
    
    MEMORY_SPI_open(SPI_MASTER_DEVICE); 

}


//************************************************************
//               is_memory_inserted  Implementation
//************************************************************
uint8_t is_memory_inserted (void){
    
    //Check if the detection switch is pressed
    if (MEMORY_DETECTION_PIN == 0x00){
        delay_ms (10);
        
        if (MEMORY_DETECTION_PIN == 0x00){
            return (MEMORY_DETECTED);
        } else {
            return (MEMORY_NOT_DETECTED);
        }
    } else {
        return (MEMORY_NOT_DETECTED);
    }
}


//************************************************************
//            blink_memory_LED  Implementation
//************************************************************
void blink_memory_LED(uint8_t time_ON) {
    LED_1_MEMORY = 0x01; 
    delay_ms(time_ON); 
    LED_1_MEMORY = 0x00;
}


//************************************************************
//            get_fattime  Implementation
//************************************************************
#ifdef MEMORY_BOARD_WITH_RTC  
DWORD get_fattime (void){
    
    static DWORD RTC_time = 0x00000000;
    
    RTC_time |= LTlib_BCD8_to_uint8 (RTCC_get_seconds());
    RTC_time |= LTlib_BCD8_to_uint8 (RTCC_get_minutes()) << 5;     
    RTC_time |= LTlib_BCD8_to_uint8 (RTCC_get_hours()) << 11;  
    RTC_time |= LTlib_BCD8_to_uint8 (RTCC_get_days()) << 16;
    RTC_time |= LTlib_BCD8_to_uint8 (RTCC_get_months()) << 21; 
    RTC_time |= (LTlib_BCD8_to_uint8 (RTCC_get_years()) + 20) << 25; 
    
    /*Format used by the FAT library
    
       bit31:25     Year origin from the 1980 (0..127, e.g. 37 for 2017)
       bit24:21     Month (1..12)
       bit20:16     Day of the month (1..31)
       bit15:11     Hour (0..23)
       bit10:5      Minute (0..59)
       bit4:0       Second / 2 (0..29, e.g. 25 for 50) */
    
}
#endif


//************************************************************
//        deactivate_memory  Implementation
//************************************************************
static void deactivate_memory (void){
    
    // Wait for card ready 
	memory_CS_enabled();		
    
	memory_CS_disabled();

    //Disable SPI
    MEMORY_SPI_close();  

    //Reset memory status 
	Stat |= STA_NOINIT;	 
}


//************************************************************
//          wait_memory_ready  Implementation
//************************************************************
static BYTE wait_memory_ready (void){
    
	BYTE res;

    //Wait for ready in timeout of 500ms 
	Timer2 = 500;	
    
	do
		res = MEMORY_SPI_read_byte();
	while ((res != 0xFF) && Timer2);

	return res;
}


//************************************************************
//        memory_CS_deselect  Implementation
//************************************************************
static void memory_CS_disabled (void){
    
    MEMORY_CS_PIN = MEMORY_CS_PIN_DISABLED;
	MEMORY_SPI_read_byte();
}


//************************************************************
//      memory_CS_enabled  Implementation
//************************************************************
static int memory_CS_enabled (void){
    
    MEMORY_CS_PIN = MEMORY_CS_PIN_ENABLED;
    
	if (wait_memory_ready() != 0xFF) {
		memory_CS_disabled();
		return 0;
	}
    
	return 1;
}


//************************************************************
//               send_cmd  Implementation
//************************************************************
static BYTE send_cmd ( BYTE cmd, DWORD arg){
    
	BYTE n, res;

	if (cmd & 0x80) {	/* ACMD<n> is the command sequence of CMD55-CMD<n> */
		cmd &= 0x7F;
		res = send_cmd(CMD55, 0);
		if (res > 1) return res;
	}

	/* Select the card and wait for ready */
	memory_CS_disabled();
	if (!memory_CS_enabled()) return 0xFF;

	/* Send command packet */
	MEMORY_SPI_write_byte(0x40 | cmd);			/* Start + Command index */
	MEMORY_SPI_write_byte((BYTE)(arg >> 24));	/* Argument[31..24] */
	MEMORY_SPI_write_byte((BYTE)(arg >> 16));	/* Argument[23..16] */
	MEMORY_SPI_write_byte((BYTE)(arg >> 8));		/* Argument[15..8] */
	MEMORY_SPI_write_byte((BYTE)arg);			/* Argument[7..0] */
	n = 0x01;						/* Dummy CRC + Stop */
	if (cmd == CMD0) n = 0x95;		/* Valid CRC for CMD0(0) */
	if (cmd == CMD8) n = 0x87;		/* Valid CRC for CMD8(0x1AA) */
	MEMORY_SPI_write_byte(n);

	/* Receive command response */
	if (cmd == CMD12) MEMORY_SPI_read_byte();	/* Skip a stuff byte when stop reading */
	n = 10;							/* Wait for a valid response in timeout of 10 attempts */
	do
		res = MEMORY_SPI_read_byte();
	while ((res & 0x80) && --n);

	return res;			/* Return with the response value */
}



//************************************************************
//               xmit_datablock  Implementation
//************************************************************
#if _READONLY == 0
static int xmit_datablock (	/* 1:OK, 0:Failed */
	const BYTE *buff,	/* 512 byte data block to be transmitted */
	BYTE token			/* Data token */
)
{
	BYTE resp;
	UINT bc = 512;


	if (wait_memory_ready() != 0xFF) return 0;

	MEMORY_SPI_write_byte(token);		/* Xmit a token */
	if (token != 0xFD) {	/* Not StopTran token */
		do {						/* Xmit the 512 byte data block to the MMC */
			MEMORY_SPI_write_byte(*buff++);
			MEMORY_SPI_write_byte(*buff++);
		} while (bc -= 2);
		MEMORY_SPI_write_byte(0xFF);				/* CRC (Dummy) */
		MEMORY_SPI_write_byte(0xFF);
		resp = MEMORY_SPI_read_byte();			/* Receive a data response */
		if ((resp & 0x1F) != 0x05)	/* If not accepted, return with error */
			return 0;
	}

	return 1;
}
#endif	/* _READONLY */



//************************************************************
//               disk_write  Implementation
//************************************************************
#if _READONLY == 0
DRESULT disk_write (
	BYTE pdrv,				/* Physical drive nmuber (0) */
	const BYTE *buff,		/* Pointer to the data to be written */
	DWORD sector,			/* Start sector number (LBA) */
	UINT count				/* Sector count (1..255) */
)
{
	if (pdrv || !count) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;
	if (Stat & STA_PROTECT) return RES_WRPRT;

	if (!(CardType & CT_BLOCK)) sector *= 512;	/* Convert to byte address if needed */

	if (count == 1) {		/* Single block write */
		if ((send_cmd(CMD24, sector) == 0)	/* WRITE_BLOCK */
			&& xmit_datablock(buff, 0xFE))
			count = 0;
	}
	else {				/* Multiple block write */
		if (CardType & CT_SDC) send_cmd(ACMD23, count);
		if (send_cmd(CMD25, sector) == 0) {	/* WRITE_MULTIPLE_BLOCK */
			do {
				if (!xmit_datablock(buff, 0xFC)) break;
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD))	/* STOP_TRAN token */
				count = 1;
		}
	}
	memory_CS_disabled();

	return count ? RES_ERROR : RES_OK;
}
#endif /* _READONLY */



//************************************************************
//               rcvr_datablock  Implementation
//************************************************************
static int rcvr_datablock (	BYTE *buff, UINT btr ) {
    
	BYTE token;

	Timer1 = 100;
	do {							/* Wait for data packet in timeout of 100ms */
		token = MEMORY_SPI_read_byte();
	} while ((token == 0xFF) && Timer1);

	if(token != 0xFE) return 0;		/* If not valid data token, return with error */

	do {	/* Receive the data block into buffer */
        
        *buff++ = MEMORY_SPI_read_byte();
        *buff++ = MEMORY_SPI_read_byte();
        *buff++ = MEMORY_SPI_read_byte();
        *buff++ = MEMORY_SPI_read_byte();
       
	} while (btr -= 4);
	MEMORY_SPI_read_byte();						/* Discard CRC */
	MEMORY_SPI_read_byte();

	return 1;						/* Return with success */
}


//******************************************************************************
//               Public Functions
//******************************************************************************


//************************************************************
//               disk_initialize  Implementation
//************************************************************
DSTATUS disk_initialize (BYTE drv ) {
    
	BYTE n, cmd, ty, ocr[4];

    //Activate the SPI module and memory
	activate_memory();							
    
    MEMORY_SPI_baudrate(MEMORY_LOW_SPEED);
    
	for (n = 80; n; n--) MEMORY_SPI_read_byte();	/* 80 dummy clocks */

	ty = 0;
	if (send_cmd(CMD0, 0) == 1) {			/* Enter Idle state */
		Timer1 = 100;						/* Initialization timeout of 1000 msec */
		if (send_cmd(CMD8, 0x1AA) == 1) {	/* SDv2? */
			for (n = 0; n < 4; n++) ocr[n] = MEMORY_SPI_read_byte();			/* Get trailing return value of R7 resp */
			if (ocr[2] == 0x01 && ocr[3] == 0xAA) {				/* The card can work at vdd range of 2.7-3.6V */
				while (Timer1 && send_cmd(ACMD41, 0x40000000));	/* Wait for leaving idle state (ACMD41 with HCS bit) */
				if (Timer1 && send_cmd(CMD58, 0) == 0) {			/* Check CCS bit in the OCR */
					for (n = 0; n < 4; n++) ocr[n] = MEMORY_SPI_read_byte();
					ty = (ocr[0] & 0x40) ? CT_SD2|CT_BLOCK : CT_SD2;	/* SDv2 */
				}
			}
		} else {							/* SDv1 or MMCv3 */
			if (send_cmd(ACMD41, 0) <= 1) 	{
				ty = CT_SD1; cmd = ACMD41;	/* SDv1 */
			} else {
				ty = CT_MMC; cmd = CMD1;	/* MMCv3 */
			}
			while (Timer1 && send_cmd(cmd, 0));		/* Wait for leaving idle state */
			if (!Timer1 || send_cmd(CMD16, 512) != 0)	/* Set read/write block length to 512 */
				ty = 0;
		}
	}
    
	CardType = ty;
	memory_CS_disabled();

	if (ty) {			/* Initialization succeded */
		Stat &= ~STA_NOINIT;	/* Clear STA_NOINIT */
        
		MEMORY_SPI_baudrate(MEMORY_HIGH_SPEED);
        
	} else {			/* Initialization failed */
		deactivate_memory();
	}

	return Stat;
}




//************************************************************
//               disk_status  Implementation
//************************************************************
DSTATUS disk_status (BYTE drv ){
    
	if (drv) return STA_NOINIT;		/* Supports only single drive */
	return Stat;
}




//************************************************************
//               disk_read  Implementation
//************************************************************
DRESULT disk_read (BYTE pdrv, BYTE* buff, DWORD sector, UINT count) {
    
	if (pdrv || !count) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;

	if (!(CardType & CT_BLOCK)) sector *= 512;	/* Convert to byte address if needed */

	if (count == 1) {		/* Single block read */
		if ((send_cmd(CMD17, sector) == 0)	/* READ_SINGLE_BLOCK */
			&& rcvr_datablock(buff, 512))
			count = 0;
	}
	else {				/* Multiple block read */
		if (send_cmd(CMD18, sector) == 0) {	/* READ_MULTIPLE_BLOCK */
			do {
				if (!rcvr_datablock(buff, 512)) break;
				buff += 512;
			} while (--count);
			send_cmd(CMD12, 0);				/* STOP_TRANSMISSION */
		}
	}
	memory_CS_disabled();

	return count ? RES_ERROR : RES_OK;
}




//************************************************************
//               w  Implementation
//************************************************************
#if _READONLY == 0
DRESULT w (
	BYTE drv,				/* Physical drive nmuber (0) */
	const BYTE *buff,		/* Pointer to the data to be written */
	DWORD sector,			/* Start sector number (LBA) */
	BYTE count				/* Sector count (1..255) */
)
{
	if (drv || !count) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;
	if (Stat & STA_PROTECT) return RES_WRPRT;

	if (!(CardType & CT_BLOCK)) sector *= 512;	/* Convert to byte address if needed */

	if (count == 1) {		/* Single block write */
		if ((send_cmd(CMD24, sector) == 0)	/* WRITE_BLOCK */
			&& xmit_datablock(buff, 0xFE))
			count = 0;
	}
	else {				/* Multiple block write */
		if (CardType & CT_SDC) send_cmd(ACMD23, count);
		if (send_cmd(CMD25, sector) == 0) {	/* WRITE_MULTIPLE_BLOCK */
			do {
				if (!xmit_datablock(buff, 0xFC)) break;
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD))	/* STOP_TRAN token */
				count = 1;
		}
	}
	memory_CS_disabled();

	return count ? RES_ERROR : RES_OK;
}
#endif /* _READONLY */


//************************************************************
//               disk_ioctl  Implementation
//************************************************************
DRESULT disk_ioctl (BYTE drv, BYTE ctrl, void *buff ) {
    
	DRESULT res;
	BYTE n, csd[16], *ptr = buff;
	DWORD csize;


	if (drv) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;

	res = RES_ERROR;
	switch (ctrl) {
		case CTRL_SYNC :	/* Flush dirty buffer if present */
			if (memory_CS_enabled()) {
				memory_CS_disabled();
				res = RES_OK;
			}
			break;

		case GET_SECTOR_COUNT :	/* Get number of sectors on the disk (WORD) */
			if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
				if ((csd[0] >> 6) == 1) {	/* SDv2? */
					csize = csd[9] + ((WORD)csd[8] << 8) + 1;
					*(DWORD*)buff = (DWORD)csize << 10;
				} else {					/* SDv1 or MMCv2 */
					n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
					csize = (csd[8] >> 6) + ((WORD)csd[7] << 2) + ((WORD)(csd[6] & 3) << 10) + 1;
					*(DWORD*)buff = (DWORD)csize << (n - 9);
				}
				res = RES_OK;
			}
			break;

		case GET_SECTOR_SIZE :	/* Get sectors on the disk (WORD) */
			*(WORD*)buff = 512;
			res = RES_OK;
			break;

		case GET_BLOCK_SIZE :	/* Get erase block size in unit of sectors (DWORD) */
			if (CardType & CT_SD2) {	/* SDv2? */
				if (send_cmd(ACMD13, 0) == 0) {		/* Read SD status */
					MEMORY_SPI_read_byte();
					if (rcvr_datablock(csd, 16)) {				/* Read partial block */
						for (n = 64 - 16; n; n--) MEMORY_SPI_read_byte();	/* Purge trailing data */
						*(DWORD*)buff = 16UL << (csd[10] >> 4);
						res = RES_OK;
					}
				}
			} else {					/* SDv1 or MMCv3 */
				if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {	/* Read CSD */
					if (CardType & CT_SD1) {	/* SDv1 */
						*(DWORD*)buff = (((csd[10] & 63) << 1) + ((WORD)(csd[11] & 128) >> 7) + 1) << ((csd[13] >> 6) - 1);
					} else {					/* MMCv3 */
						*(DWORD*)buff = ((WORD)((csd[10] & 124) >> 2) + 1) * (((csd[11] & 3) << 3) + ((csd[11] & 224) >> 5) + 1);
					}
					res = RES_OK;
				}
			}
			break;

		case MMC_GET_TYPE :		/* Get card type flags (1 byte) */
			*ptr = CardType;
			res = RES_OK;
			break;

		case MMC_GET_CSD :	/* Receive CSD as a data block (16 bytes) */
			if ((send_cmd(CMD9, 0) == 0)	/* READ_CSD */
				&& rcvr_datablock(buff, 16))
				res = RES_OK;
			break;

		case MMC_GET_CID :	/* Receive CID as a data block (16 bytes) */
			if ((send_cmd(CMD10, 0) == 0)	/* READ_CID */
				&& rcvr_datablock(buff, 16))
				res = RES_OK;
			break;

		case MMC_GET_OCR :	/* Receive OCR as an R3 resp (4 bytes) */
			if (send_cmd(CMD58, 0) == 0) {	/* READ_OCR */
				for (n = 0; n < 4; n++)
					*((BYTE*)buff+n) = MEMORY_SPI_read_byte();
				res = RES_OK;
			}
			break;

		case MMC_GET_SDSTAT :	/* Receive SD status as a data block (64 bytes) */
			if (send_cmd(ACMD13, 0) == 0) {	/* SD_STATUS */
				MEMORY_SPI_read_byte();
				if (rcvr_datablock(buff, 64))
					res = RES_OK;
			}
			break;

		default:
			res = RES_PARERR;
	}

	memory_CS_disabled();

	return res;
}

